// Declaration of the global object
let globalData = {
  accountingExercises: null,
  accountingEntries: null,
  salesInvoices: null,
  customers: null,
  accountingAccounts: null,
  products: null
};

//Check if the company contains any data and ask the user if they want to import sample data if it does not.
function askForImportData(userConfirmedWithoutDeletingFiscalYears = false) {
  qm.queryCheckOrganizationData.isImportFromTemplatesRunning = true;
  genericFetchPublicAPI(headerWithOrganization, qm.queryCheckOrganizationData, null, (data) => handleResponse(data, userConfirmedWithoutDeletingFiscalYears));
}

function handleResponse(data, userConfirmedWithoutDeletingFiscalYears = false) {
  if (data && data.data) {
    globalData = {
      accountingExercises: data.data.accountingExercises,
      accountingEntries: data.data.accountingEntries,
      salesInvoices: data.data.salesInvoices,
      customers: data.data.customers,
      accountingAccounts: data.data.accountingAccounts,
      products: data.data.products
    };

    let logMessage = '';
    for (let key in globalData) {
      logMessage += `${key}: ${globalData[key] && globalData[key].edges ? globalData[key].edges.length : 0}\n`;
    }   
    console.log(logMessage);

    if (areEntitiesEmpty([globalData.customers])) {
      if (!areEntitiesEmpty([globalData.accountingExercises])) {
        let startDate = globalData.accountingExercises.edges[0].node.startDate.substr(0, 10);
        let janFirstYear = new Date(Date.UTC(new Date().getFullYear(), 0, 1)).toISOString().substring(0, 10);
        let janFirstPreviousYear = new Date(Date.UTC(new Date().getFullYear() - 1, 0, 1)).toISOString().substring(0, 10);

        console.log("INFO : Number of fiscal years: " + globalData.accountingExercises.edges.length + " -  First Exercise startDate: " + startDate);

        if (globalData.accountingExercises.edges.length > 1 || (startDate != janFirstPreviousYear && startDate != janFirstYear)) {
          let message = `Your business can be enhanced by Demo Data.<br><br>`;
          const criticalPart = `But it already contains fiscal years that are not compatible with the data to be imported.<br><br>
          If you wish to import demo data:<br><ul><li>please first go to Sage Active/Settings/Fiscal Years and <strong>delete the fiscal years</strong>,</li><li>then click the <strong>Yes</strong> button.</li></ul>`;
        
            // If userConfirmedWithoutDeletingFiscalYears is true, display the critical part in red
          if (userConfirmedWithoutDeletingFiscalYears) {
            message += `<span style="color: red;">${criticalPart}</span>`;
          } else {
            message += criticalPart;
          }
        
          dlgBoxYesNo("info", "Import Demo Data", message, (choice) => handleImportDeleteExercises(choice));
          return;
        }
        
      }

      checkPermissionsAndAct(['createCustomer'],
        "Your company does not contain any data, but you do not have the necessary permissions to import demonstration data.",
        askForImportDemo); 
    } else if (areEntitiesEmpty([globalData.salesInvoices])) {
      checkPermissionsAndAct(['createSalesInvoice'],
        "Your company does not contain any sales Invoices, but you do not have the necessary permissions to import sales invoices.",
        askForImportSales);
    }
  }
}

// Check if all non-null entities are empty
function areEntitiesEmpty(entities) {
  return entities.filter(entity => entity !== null).every(entity => entity.edges.length === 0);
}
// Check if all non-null entities are empty
function areEntitiesEmpty(entities) {
  if (!entities || !Array.isArray(entities)) {
    return true; 
  }
  return entities.filter(entity => entity !== null).every(entity => entity.edges && entity.edges.length === 0);
}

// Check if the user has the required permissions and act accordingly
function checkPermissionsAndAct(requiredPolicies, warningMessage, actionFunction) {
  const missingPolicy = requiredPolicies.find(policy => !(policy in userPolicies));

  if (missingPolicy) {
    setTimeout(() => { displayBusinessError(`Missing policy: ${missingPolicy} `); }, 1000);
    return;
  }

  if (!requiredPolicies.every(policy => userPolicies[policy])) {
    setTimeout(() => { displayBusinessWarning(warningMessage); }, 1000);
    return;
  }

  actionFunction();
}

// Ask the user if they want to import sample sales documents
function askForImportSales() {
  const message = "Your company does not contain any Demo Sales Invoices.<br><br>Would you like to import Demo Sales Documents?";
  let configs = [salesQuotesConfig,salesOrdersConfig,salesDeliveryNotesConfig,salesInvoicesConfig]; 
  if (areEntitiesEmpty([globalData.products])) configs.unshift(productConfig); 
  console.log(configs.map(config => config.objectName || 'Unknown config'));
  askForImport("Import Demo Sales documents", message, () => importFromTemplates(configs));
}

// Ask the user if they want to import sample sales documents
function askForImportSalesInvoices() {
  const message = "Would you like to import Sales Invoices Demo Data?";
  let configs = [salesInvoicesConfig]; 
  if (areEntitiesEmpty([globalData.products])) configs.unshift(productConfig); 
  console.log(configs.map(config => config.objectName || 'Unknown config'));
  askForImport("Import Demo Sales Invoices", message, () => importFromTemplates(configs));
}

// Ask the user if they want to import demonstration data
function askForImportDemo() {
  const message = "Your company does not contain any Demo Data.<br><br>Would you like to import Demo Data?";
  
  // Check if globalData.accountingAccounts is empty before including importAccountingPlanMasterConfig
  let configs = [productConfig, accountingExerciseConfig, customerConfig, supplierConfig, employeeConfig, accountingEntriesConfig,salesQuotesConfig,salesOrdersConfig,salesDeliveryNotesConfig,salesInvoicesConfig]; 
  if (areEntitiesEmpty([globalData.accountingAccounts])) {
    configs.unshift(importAccountingPlanMasterConfig);  
  }
  console.log(configs.map(config => config.objectName || 'Unknown config'));
  askForImport("Import Demo Data", message, () => importFromTemplates(configs));
}

// Ask the user if they want to import accountingEntries data
function askForImportAccountingEntries() {
  askForImportEntity(accountingEntriesConfig);
}
// Ask the user if they want to import products
function askForImportProducts() {
   askForImportEntity(productConfig);
}


// Ask the user if they want to import something
function askForImportEntity(config) {
  const name=capitalizeFirstLetter(config.objectName);
  const message = `Would you like to import ${name} Demo Data?`;
  askForImport(`Import ${name} Demo Data`, message, () => importFromTemplates([config]));
}
function capitalizeFirstLetter(word) {
  return word.charAt(0).toUpperCase() + word.slice(1).toLowerCase();
}


// standard function to propose an import
function askForImport(dialogTitle, dialogMessage, importFunction) {
   dlgBoxYesNo("info", dialogTitle, dialogMessage, (choice) => handleImportAction(choice, importFunction));
}

// Handle the user's choice to import data
function handleImportAction(choice, importFunction) {
  if (choice) {
    dlgBoxYesNoCancel();
    disableInterface(); 
    importFunction();
  } else {
    dlgBoxYesNoCancel();
  }
}
function handleImportDeleteExercises(choice) {
  if (choice) {
    dlgBoxYesNoCancel();
    disableInterface(); 
    setTimeout(() => { askForImportData(true); }, 1000);  
  } else {
    dlgBoxYesNoCancel();
  }
}




