var clientId, sbcAuthUrl ;
var currentAccessToken;
var currentRefreshToken;

var allowedSageActiveEnvs; //  allowed Sage active environments
var sbcAuthEnvBySageActiveEnv; // map SBC Auth environments to Sage Active environments
var sbcAuthConfigBySageActiveEnvs; // hold the configuration SBCAuth clientId & SBCAuth Url for for each Sage Active environment


//Checks for an existing access token 
//Or initiates the authentication flow : step 1 authorizationRequest, then step 2 completeAuthCodeExchange 
async function checkAccessToken() {

  //Ask the proxy to return info about allowed environments
  try {
    const data = await fetchAllowedEnvironmentsAndSBCAuthConfiguration();
    allowedSageActiveEnvs = data.allowedSageActiveEnvs;
    sbcAuthEnvBySageActiveEnv=data.sbcAuthEnvBySageActiveEnv;
    sbcAuthConfigBySageActiveEnvs=data.sbcAuthConfigBySageActiveEnvs;

    //Retrieve the current environment, if no exists use the first available environment and redirect to login
    sageActiveEnvironment = localStorage.getItem("sageActiveEnvironment");
    if (!sageActiveEnvironment) {
      sageActiveEnvironment = allowedSageActiveEnvs[0];
      sessionStorage.setItem('isLoggingOut',true);
      localStorage.setItem("sageActiveEnvironment", sageActiveEnvironment);
    }
    // Remove the old environment from the title if it exists and Add the new environnement to the title
    document.title = document.title.replace(/\(\w+\)/, '')+ ` (${sageActiveEnvironment.toUpperCase()})`;

  } catch (error) {
    displayBusinessError('Error fetching environments:'+error);
    return; 
  }

  //Get the Client Id and authentication url for the current environment
  const config=getAuthConfig(sageActiveEnvironment);
  clientId=config.clientId;
  sbcAuthUrl=config.sbcAuthUrl;
  
  //If redirection to logout
  if (sessionStorage.getItem('isLoggingOut') === 'true') {
    sessionStorage.removeItem('isLoggingOut');
    login();
    return;
  }
  
  //Get the current token & refresh token
  currentAccessToken=sessionStorage.getItem("currentAccessToken");
  currentRefreshToken=sessionStorage.getItem("currentRefreshToken");

  //if new authentication required because the current token is empty.
  if (!currentAccessToken){
    var code= new URL(window.location.href).searchParams.get('code');
    //Step 1 
    if (!code){
        //If error during the last authentication, propose to reconnect to the default environment instead of the current
        if (sessionStorage.getItem('codeVerifier')){
          sessionStorage.removeItem('codeVerifier');
          displayBusinessError(`Authentication failed, verify the Client ID for the ${sageActiveEnvironment.toUpperCase()} environment.`);
          login();
          return;
        }
        await authorizationRequest();
    }
    //Step 2
    else{
        await completeAuthCodeExchange(code); 
    }
  }
}

//Authorization request
async function authorizationRequest(){
    
    function generateRandomString() {
        var array = new Uint32Array(28);
        window.crypto.getRandomValues(array);
        return Array.from(array, dec => ('0' + dec.toString(16)).substr(-2)).join('');
    }
    var codeVerifier = generateRandomString();
    sessionStorage.setItem('codeVerifier', codeVerifier);
    async function sha256(s) {
        const digest = await crypto.subtle.digest({ name: "SHA-256" }, new TextEncoder().encode(s));
        return btoa(String.fromCharCode(...new Uint8Array(digest))).replace(/=/g, '').replace(/\+/g, '-').replace(/\//g, '_');  
    }
    
    var codeChallenge = await sha256(codeVerifier);

    var authorizationEndpoint = sbcAuthUrl+"/connect/authorize";
    var redirectUri = encodeURIComponent(location.href);
    var scopes=encodeURIComponent("RDSA WDSA offline_access");   
    var state="1234";

    var url = `${authorizationEndpoint}?client_id=${clientId}&response_type=code&scope=${scopes}&redirect_uri=${redirectUri}&code_challenge=${codeChallenge}&code_challenge_method=S256&state=${state}`;
    window.location.href = url;
}

//Resend authorization code to get access token
async function completeAuthCodeExchange(code){
    var redirectUri = encodeURIComponent(window.location.protocol + '//' + window.location.host + window.location.pathname);
    var codeVerifier = sessionStorage.getItem('codeVerifier');
    sessionStorage.removeItem('codeVerifier');
    await fetch(sbcAuthUrl+"/connect/token", {
        method: 'POST',
        body: `grant_type=authorization_code&code=${code}&redirect_uri=${redirectUri}&client_id=${clientId}&code_verifier=${codeVerifier}`,
        headers: {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
    })
    .then(response => response.json())
    .then(data => {
        currentAccessToken=data.access_token;
        currentRefreshToken=data.refresh_token;
        sessionStorage.setItem("currentAccessToken",currentAccessToken);
        sessionStorage.setItem("currentRefreshToken",currentRefreshToken);
    })
    history.pushState({}, "", window.location.origin + window.location.pathname);
}

function logout(forceLogout=true){ 
  sessionStorage.removeItem('currentAccessToken');
  sessionStorage.removeItem('currentRefreshToken');
  sessionStorage.setItem('isLoggingOut', 'true');
  if (forceLogout && clientId){
      waitInfo(false);
      hideBusinessMessages();
      disableInterface();
      $("#logoutInfoTitle").html(`${sageActiveEnvironment} Sage Active Environment Logout Process</b>.`);
      $("#logoutInfoEnv").html(`A new tab will be opened in your browser to execute the user logout.`);
      $('#logoutInfo').show();
  } else {
     redirectToLogin();
  }
}

function logoutExecute(){
  $('#logoutInfo').hide();
  $('#logoutInProgress').show();
  waitInfo();
  displayBusinessInfo('Waiting for the <strong>Welcome</strong> Sage Active tab to close',false);
  const domain = ['stagefr','stagees','stagede', 'fr', 'es', 'de'].includes(sageActiveEnvironment.toLowerCase()) ? "active.sage" : "sageeurekadev";
  const logoutTab = window.open(`https://www.${sageActiveEnvironment.toLowerCase()}.${domain}.com/logout`, '_blank');

  const checkIfClosed = setInterval(() => {
    if (logoutTab.closed) {
      $('#logoutInProgress').hide();
      clearInterval(checkIfClosed); 
      displayBusinessInfo("You have been logged out successfully.");
      var returnTo = window.location.protocol + '//' + window.location.host + window.location.pathname;
      const logoutUrl = `${sbcAuthUrl}/account/logout?client_id=${clientId}&returnTo=${encodeURIComponent(returnTo)}`;
      window.location.href = logoutUrl;
    }
  }, 1000);
}


//renew a Refresh Token 
function refreshToken() {
  return new Promise((resolve, reject) => {
    var tokenRenewalUrl = sbcAuthUrl + "/connect/token";
    var body = `grant_type=refresh_token&refresh_token=${encodeURIComponent(currentRefreshToken)}&client_id=${clientId}`;
    
    fetch(tokenRenewalUrl, { 
      method: "POST",
      headers: {
        'Content-Type': 'application/x-www-form-urlencoded'
      },
      body: body
    })
    .then(response => response.json())
    .then(data => {
      if (data.access_token && data.refresh_token) {
        currentAccessToken=data.access_token;
        currentRefreshToken=data.refresh_token;
        sessionStorage.setItem("currentAccessToken",currentAccessToken);
        sessionStorage.setItem("currentRefreshToken",currentRefreshToken);
        resolve(data.access_token);
      } else {
        reject(new Error('Failed to refresh token.'));
      }
    })
    .catch(error => {
      displayBusinessError('401');
      reject(error);
    });
  });
}

//Requests the proxy for the allowed Sage Active environment contexts, the SBCAuth/Environment mappings, and the url and clientId configuration of each SBCAuth app.
async function fetchAllowedEnvironmentsAndSBCAuthConfiguration() {
  const response = await fetch(urlAPI + "/environments", { headers: { 'X-FrontEnd-URL': window.location.href } });
  if (!response.ok) {
    return response.text().then(text => {
      throw new Error(`HTTP error! status: ${response.status}, message: ${text}`);
    });
  }
  return await response.json();
}

//retrieve the clientId and SBCAuth Url for the current environment 
function getAuthConfig(sageActiveEnvironment) {
  sageActiveEnvironment = /^std/i.test(sageActiveEnvironment.toLowerCase()) && sageActiveEnvironment.toLowerCase() !== "stdkeepit" ? "standalone" : sageActiveEnvironment.toLowerCase();
  const sbcAuthEnv = sbcAuthEnvBySageActiveEnv[sageActiveEnvironment] || 'dev';
  if (sbcAuthConfigBySageActiveEnvs.hasOwnProperty(sbcAuthEnv)) {
    return sbcAuthConfigBySageActiveEnvs[sbcAuthEnv];
  } else {
    sageActiveEnvironment=allowedSageActiveEnvs[0];
    localStorage.removeItem("sageActiveEnvironment");
    return sbcAuthConfigBySageActiveEnvs[sageActiveEnvironment.toLowerCase()];
  }
}

//Propose to log to an environment
function login(){
  updateLoginVisibility();
  waitInfo(false);
  disableInterface();
  $('#login').show();
}

//Propose to log into the first Environment in addition to the current environment, if it is different from the first one
function updateLoginVisibility() {
  $("#loginCurrentEnv").html(`Login to ${sageActiveEnvironment.toUpperCase()} environment`);
  if (allowedSageActiveEnvs.length>1) {
      $("#loginTitle").html(`Choose your environment`);
      $("#changeEnv").html(`Choose another environment`);
      $("#changeEnv").show();
  } else {
      $("#loginTitle").html(`Logging into your environment`);
      $("#changeEnv").hide();
  }
}

// When the user requests to switch environments
function ChangeEnvAndForceLogout({ env="", forceLogout = true }){
  $(".logout").prop("disabled","true");
  localStorage.setItem("sageActiveEnvironment", env);

  waitInfo();
  logout(forceLogout);
}


// Redirect to the desired environment
function redirectToLogin() {
  window.location.href = window.location.href;
}

// Suggest to change environment by displaying the list of other environments
function suggestOtherEnvironment(){
  $('#login').hide();
  enableInterface();
  displayBusinessInfo("Choose the desired environment from the list");
  otherEnvs({ returnToOrganizations: false, forceLogout: false });
}