let currentCount = 0;
let maxAttempt = 5;

function createObjectFromTemplate(objectConfig, objectData, total, preImportResult,attempt = 1) {
    return new Promise((resolve, reject) => {
        const { qmObject, buildValues, postImportAfterEachRecordCreation } = objectConfig;
        const values = buildValues(objectData,preImportResult);
        var message = `Import ${objectConfig.objectName} : ${objectConfig.describe(objectData)}`;
        qmObject.isImportFromTemplatesRunning = true;
        genericFetchPublicAPI(headerWithOrganization, qmObject, values, (response) => {
            if (attempt==1) currentCount += 1;
            const mutationResultKey = response.data && Object.keys(response.data).length > 0 ? Object.keys(response.data)[0] : null;
            if (response.data && mutationResultKey && response.data[mutationResultKey] !== null) {
                const responseData = response.data[mutationResultKey];  
                displayBusinessInfo(`${currentCount}/${total} - ${message} -> Created`, "", false);

                // Call postImportAfterEachRecordCreation if defined
                if (postImportAfterEachRecordCreation && typeof postImportAfterEachRecordCreation === 'function') {
                    postImportAfterEachRecordCreation(responseData,objectData,values).then(() => {
                        resolve(response.data);
                    }).catch(err => {
                        displayBusinessError(`Post import action failed: ${err}`);
                        resolve(response.data);  // Do not block creation in case of postImport error
                    });
                } else {
                    resolve(response.data);
                }
            } else if (response.errors && response.errors.length > 0) {
                const errorMessage = response.errors[0].message;
                if (errorMessage.includes("codeAlreadyExistsInOrganization") || errorMessage.includes("accountingExerciseFollowingExerciseMustBeContiguous") || errorMessage.includes("accountingAccountsAlreadyExistInOrganization")) {
                    displayBusinessWarning(`${currentCount}/${total} - ${message} -> Already exists`, "", false);
                    resolve();
                } else {
                    displayBusinessError(`${currentCount}/${total} - ${message} -> Error : ${errorMessage}`, "", false);
                    resolve();
                }
                //CHCOP 1500 instead of 500 to deactivate the retry mechanism for 500
            } else if ((response.status === 1500 || response.status === 504) && attempt < maxAttempt) {
                displayBusinessError(`${currentCount}/${total} - ${message} -> Error : ${attempt}/${maxAttempt} -> ${response.status}`, "", false);
                setTimeout(() => createObjectFromTemplate(objectConfig, objectData, total, preImportResult,attempt + 1).then(resolve).catch(reject), 2000);
            } else {
                const errorGeneral = "Operation failed or returned no data";
                displayBusinessError(`${message} -> ${errorGeneral}`);
                resolve();
            }
        });
    });
}

// Get the template data for an entity
async function getTemplateData(objectConfig) {
    const templateName = `importdata/${activeLegislationCode}_template_${objectConfig.objectName.toLowerCase()}.json`;
    return new Promise((resolve) => {
        $.getJSON(templateName).done(function(data) {
            if (data && data.data && data.data[objectConfig.objectName] && Array.isArray(data.data[objectConfig.objectName].edges)) {
                resolve(data.data[objectConfig.objectName].edges);
            } else {
                displayBusinessError(`Invalid ${templateName} structure`);
            }
        })
        .fail(function(error) {
            displayBusinessError(`Failed to load ${templateName}: ${error.statusText || error}`);
        });
    });
}

//Read all records of a given template and call the createObjectFromTemplate function for each record. Except if the user does not have the necessary permissions
function importFromTemplate(objectConfig,preImportResult,objects) {
    if (userPolicies[objectConfig.qmObject.policy] === false) return Promise.resolve();;
    currentCount = 0;
    let parallelismLimit = (forceMaxParallelismLimit) ? maxParallelismLimit : Math.min(maxParallelismLimit,objectConfig.parallelismLimit) || maxParallelismLimit;
    displayBusinessInfo(`Importing <strong>${objectConfig.objectName}</strong> [parallelism limit=<strong>${parallelismLimit}</strong>]`, false);

    return new Promise((outerResolve, outerReject) => {
        const processInBatches = async (objects, batchSize) => {
        for (let i = 0; i < objects.length; i += batchSize) {
            const batch = objects.slice(i, i + batchSize);
            await Promise.all(batch.map(object => 
                createObjectFromTemplate(objectConfig, object.node, objects.length,preImportResult)
            ));
        }
    };

    //Restrict the number of accounting entries to import
    var objectsToImport=objects;
    if(objectConfig.objectName == "accountingEntries" && maxAccountingEntries>0){
        objectsToImport= objects.slice(0,maxAccountingEntries);
    }

    processInBatches(objectsToImport, parallelismLimit)
        .then(() => {
            displayBusinessInfo(`${objectConfig.objectName} import completed.`, false);
            outerResolve();
        })
        .catch(error => {
            displayBusinessError(`An error occurred during the import of the ${objectConfig.objectName} : ${error}.`, "", false);
            outerReject(error);
        });
    });
}

// Import all data from templates
async function importFromTemplates(configs,inMemoryNodes) {
    disableInterface();
    clearBusinessMessageHistory();
    showBusinessMessageHistory();
    startTimer();

    const importPromises = configs.map(config => async () => {
        let preImportResult = null;
        if (config.preImport) {
            preImportResult = await config.preImport();
        }
        var objects;
        if (!inMemoryNodes)
            objects = await getTemplateData(config);
        else
            objects = inMemoryNodes;

        return importFromTemplate(config, preImportResult,objects);
    });
  
    importPromises.reduce((promiseChain, currentTask) => {
      return promiseChain.then(currentTask);
    }, Promise.resolve())
    .then(() => {
      stopTimer();
      displayBusinessInfo("Import completed.");
      resetCustomer();
    })
    .catch((error) => {
        stopTimer();
    });
}

/********************************************************************/
/* Configurations of each entity to import                          */
/********************************************************************/

// Configuration to import the accounting plan master
const importAccountingPlanMasterConfig = {
    objectName: "AccountingPlanMasters",
    qmObject: qm.importAccountingPlanMaster,
    buildValues: (accountingPlanMaster) => ({
        "values": {
            "accountingPlanMasterId": accountingPlanMaster.accountingPlanMasterId,
            "subAccountLength": accountingPlanMaster.subAccountLength
        }
    }),
    describe: (accountingPlanMaster) => `Accounting Plan Master ID ${accountingPlanMaster.accountingPlanMasterId} with Sub Account Length ${accountingPlanMaster.subAccountLength}`
};

// Configuration to create an exercise
const accountingExerciseConfig = {
    objectName: "accountingExercises",
    qmObject: qm.createAccountingExercise,
    buildValues: (exercise) => ({
        "values": {
            "exercise": exercise.exercise,
            "description": exercise.description,
            "startDate": exercise.startDate,
            "numberPeriods": exercise.numberPeriods,
        }
    }),
    describe: (exercise) => `${exercise.exercise} ${exercise.description}`
};

// Configuration to create a product
const productConfig = {
    objectName: "products",
    qmObject: qm.createProduct,
    buildValues: (product) => ({
        "values": {
            "code": product.code,
            "lineDescription": product.lineDescription,
            "comments": product.comments,
            "name": product.name,
            "category": product.category,
            "salesUnitPrice": product.salesUnitPrice,
        }
    }),
    describe: (product) => `${product.code} ${product.name}`
};

// Configuration to create a customer or supplier
const customerConfig = createThirdConfig("customers", qm.createCustomer);
const supplierConfig = createThirdConfig("suppliers", qm.createSupplier);
function createThirdConfig(objectName, qmObject) {
    return {
        objectName: objectName,
        qmObject: qmObject,
        get parallelismLimit() {
            return activeLegislationCode == "FR" ? maxParallelismLimit : 1;
        },
        buildValues: (third) => ({
            "values": {
                "socialName": third.socialName,
                "tradeName": third.tradeName,
                "code": third.code,
                "countryAcronym": third.countryAcronym,
                "documentId": third.documentId,
                "vatNumber": third.vatNumber,
                ...(third.hasEquivalenceSurcharge !== null && { "hasEquivalenceSurcharge": third.hasEquivalenceSurcharge }),      
                "addresses": third.addresses.map(address => ({
                    "firstLine": address.firstLine,
                    "secondLine": address.secondLine,
                    "city": address.city,
                    "zipCode": address.zipCode,
                    "countryName": address.countryName,
                    "province": address.province
                })),
                "contacts": third.contacts.map((contact, index) => ({
                    "isDefault": index === 0,
                    "courtesy": contact.courtesy,
                    "name": contact.name,
                    "surname": contact.surname,
                    "emails": contact.emails.map((email,emailIndex) => ({
                        "isDefault": emailIndex === 0,                      
                        "emailAddress": email.emailAddress,
                        "usage": email.usage
                    })),
                    "phones": contact.phones.map((phone, phoneIndex) => ({
                        "isDefault": phoneIndex === 0,
                        "number": phone.number,
                        "type": phone.type
                    })),
                    "socialMedias": contact.socialMedias.map(socialMedia => ({
                        "link": socialMedia.link,
                        "name": socialMedia.name
                    }))
                }))
            }
        }),
        describe: (entity) => `${entity.code} ${entity.tradeName}`
    };
}

// Configuration to create an employee
const employeeConfig= {
        objectName: "employees",
        qmObject: qm.createEmployee,
        buildValues: (employee) => ({
            "values": {
                "name": employee.name,
                "shortName": employee.shortName,
                "code": employee.code,
                "addresses": employee.addresses.map(address => ({
                    "firstLine": address.firstLine,
                    "secondLine": address.secondLine,
                    "city": address.city,
                    "zipCode": address.zipCode,
                    "countryName": address.countryName
                })),
                "contacts": employee.contacts.map((contact, index) => ({
                    "isDefault": index === 0,
                    "courtesy": contact.courtesy,
                    "name": contact.name,
                    "surname": contact.surname,
                    "emails": contact.emails.map((email,emailIndex) => ({
                        "isDefault": emailIndex === 0,                      
                        "emailAddress": email.emailAddress,
                        "usage": email.usage
                    })),
                    "phones": contact.phones.map((phone, phoneIndex) => ({
                        "isDefault": phoneIndex === 0,
                        "number": phone.number,
                        "type": phone.type
                    })),
                    "socialMedias": contact.socialMedias.map(socialMedia => ({
                        "link": socialMedia.link,
                        "name": socialMedia.name
                    }))
                }))
            }
        }),
        describe: (entity) => `${entity.code} ${entity.shortName}`
}

// Configuration to create an accounting entry
const accountingEntriesConfig = {
    objectName: "accountingEntries",
    qmObject: qm.createAccountingEntry,
    parallelismLimit: 20,
    preImport: async () => await readSubAccountingLength(activeOrganizationId),
    buildValues: (entry) => {
        const externalInvoiceNumber = entry.accountingEntryLines
            .map(line => line.accountingEntryInvoice ? line.accountingEntryInvoice.externalInvoiceNumber : null)
            .find(invoiceNumber => invoiceNumber !== null);

        let description = entry.description;
        if (!description && entry.accountingEntryLines.length > 0) {
            description = entry.accountingEntryLines[0].description;
        }

        return {
            "values": {
                "date": entry.date,
                "documentNumber": externalInvoiceNumber || entry.documentNumber || "", 
                "description":description || "",
                "journalTypeCode": entry.journalType.code,
                "accountingEntryLines": entry.accountingEntryLines.map(line => ({
                    "subAccountCode": adaptSubAccountLength(line.subAccount.code),
                    "description": line.description,
                    "debitAmount": line.debitAmount,
                    "creditAmount": line.creditAmount,
                    "thirdCode": line.accountingEntryThirdParty ? line.accountingEntryThirdParty.code : null,
                }))
            }
        };
        // Format the accounting account code to the required length
        function adaptSubAccountLength(subAccount){
            if (subAccount.length == subAccountingLength) return subAccount;
            return subAccount.length > subAccountingLength ? subAccount.substring(0, subAccountingLength) : subAccount.padEnd(subAccountingLength, '0');
        }
    },
    describe: (entry) => `Entry journal ${entry.journalType.code} ${entry.date.split('T')[0]}`
};


// Configurations to create sales invoices and close or post following the status
const salesInvoicesConfig = createSalesConfig("salesInvoices", qm.createSalesInvoice, async (responseData, entry,values) => { 
    await performPostCreationActionsForInvoices(responseData, entry,values);
},1);  

//Configurations to create other sales documents
const salesQuotesConfig = createSalesConfig("salesQuotes", qm.createSalesQuote);
const salesOrdersConfig = createSalesConfig("salesOrders", qm.createSalesOrder);
const salesDeliveryNotesConfig = createSalesConfig("salesDeliveryNotes", qm.createSalesDeliveryNote);

//Configuration to create a common sales document
function createSalesConfig(objectName, qmObject,postImportAfterEachRecordCreation,parallelismLimit) {
    return {
        objectName: objectName,
        qmObject: qmObject,
        parallelismLimit: parallelismLimit || 3,
        preImport: async () => {
            if (objectName === "salesInvoices")  await fetchSalesJournalId();
            return await processGetCustomerIdsAndProductIds({ objectName });
        },
        postImportAfterEachRecordCreation,
        buildValues: (entry, { customersDict, productsDict }) => {
            //CHCOP force the first customer if not found while quotes with leads are not supported
            const customerId = entry.customer && customersDict[entry.customer.code] ? customersDict[entry.customer.code] : Object.values(customersDict)[0];
            
            const values= {
                "documentDate": entry.documentDate,
                "operationalNumber": entry.operationalNumber || "",
                "customerId": customerId || "",
                "lines": entry.lines.map(line => ({
                    "productId": line.productCode && productsDict[line.productCode] ? productsDict[line.productCode] : null,
                    "totalQuantity": line.totalQuantity,
                    "unitPrice": line.unitPrice,
                    "firstDiscount": line.firstDiscount || 0
                }))
            };
            if (objectName=="salesInvoices") values.salesOrderNumber = entry.salesOrderNumber || "";
            
            return { "values": values };
        },
        describe: (entry) => `${objectName.slice(0, -1)} ${entry.documentDate.split('T')[0]} ${entry.status} for customer ${entry.customer ? entry.customer.code : "unknown"}`
    };
}


/********************************************************************/
/* preimport functions                                              */
/********************************************************************/

// Retrieve customerId & ProductId from the codes in the template
async function processGetCustomerIdsAndProductIds(salesInvoicesConfig) {
    const templateData = await getTemplateData(salesInvoicesConfig);  
    
    const customerCodes = new Set();
    const productCodes = new Set();
    
    templateData.forEach(({ node: invoice }) => {
        if (invoice.customer && invoice.customer.code) {
            customerCodes.add(invoice.customer.code);
        }
        invoice.lines.forEach((line) => {
            if (line.productCode) {
                productCodes.add(line.productCode);
            }
        });
    });
    
    const customersDict = await fetchIdsForCodes([...customerCodes], 'customers');
    const productsDict = await fetchIdsForCodes([...productCodes], 'products');
    return { customersDict, productsDict };
}

// Function to fetch the ids for the given codes
function fetchIdsForCodes(codes, entityType) {
    return new Promise((resolve, reject) => {
      if (codes.length === 0) {
        resolve({}); 
        return;
      }
  
      let query="";
      let whereClause="";
    
      if (entityType === 'customers') {
        query = qm.queryCustomersIdByCode;
        whereClause = `where: { code: { in: [${codes.map(code => `"${code}"`).join(", ")}] } }`;
      } else if (entityType === 'products') {
        query = qm.queryProductsIdByCode;
        whereClause = `where: { code: { in: [${codes.map(code => `"${code}"`).join(", ")}] } }`;
      }
  
      genericFetchPublicAPI(headerWithOrganization, query, null, handleResponse,whereClause);
  
      function handleResponse(data) {
        if (data && data.data && data.data[entityType] && Array.isArray(data.data[entityType].edges)) {
          const idsDict = {};
          data.data[entityType].edges.forEach(entity => {
            idsDict[entity.node.code] = entity.node.id;
          });
          resolve(idsDict);
        }
      }
    });
}

//Return the Length of accounts
var subAccountingLength;
function readSubAccountingLength(orgId) {
    return new Promise((resolve) => {
      genericFetchPublicAPI(headerWithOrganization, qm.queryOrganizationAccountingSetupByOrgId, { id: orgId }, handleResponse);
  
      function handleResponse(data) {
        if (data && data.data && data.data.organizationAccountingSetupByOrgId) {
          subAccountingLength=data.data.organizationAccountingSetupByOrgId.subAccountingLength;
          resolve();
        } 
      }
    });
  }
  
  //Return the sales journal id
  var salesJournalId;
  function fetchSalesJournalId() {
    return new Promise((resolve) => {
        genericFetchPublicAPI(headerWithOrganization, qm.querySalesJournalId, null, handleResponse);
        function handleResponse(data) {
            if (data && data.data && data.data.journalTypes && data.data.journalTypes.edges.length > 0) {
              salesJournalId = data.data.journalTypes.edges[0].node.id;
                resolve();
            }
        }
    });
  }

/********************************************************************/
/* postImportAfterEachRecordCreation functions                      */
/********************************************************************/

// Depending on the status of the sales invoice, close, or close and post it  
async function performPostCreationActionsForInvoices(responseData, entry) {
    switch (entry.status) {
        case "Closed":
            await closeSalesInvoice(responseData.id,entry);
            break;

        case "Posted":
        case "PartiallyCollected":
        case "Collected":
        case "PartiallyOverdue":
        case "Overdue":
            var responseDataClose= await closeSalesInvoice(responseData.id,entry);
            await postSalesInvoice(responseDataClose, entry);
            break;

        case "Nothing":
        default:
            break;
    }
}

// Close a sales invoice   
async function closeSalesInvoice (invoiceId,entry) {
    return new Promise((resolve) => {
        genericFetchPublicAPI(headerWithOrganization, qm.closeSalesInvoice, { "input": { "salesInvoiceId": invoiceId } }, handleResponse);
        function handleResponse(data) {
            if (data && data.data && data.data.closeSalesInvoice) {
                displayBusinessInfo(`Sales Invoice ${entry.documentDate.split('T')[0]} ${entry.status} for customer ${entry.customer ? entry.customer.code : "unknown"} is closed`,false);
                resolve(data.data.closeSalesInvoice);
            }
            else{
                resolve();
            }
        }
    });
}

//Post a sales invoice
async function postSalesInvoice(responseDataClose,entry) {
    return new Promise((resolve) => {
        const invoiceLabel = activeLegislationCode === 'FR' ? 'Facture de vente' : activeLegislationCode === 'ES' ? 'Factura de venta' : activeLegislationCode === 'DE' ? 'Verkaufsrechnung' : 'Sales Invoice';
        const queryVariables = {
            "input": {
                "salesInvoiceId": responseDataClose.id,
                "accountingEntryDescription": `${invoiceLabel} : ${responseDataClose.operationalNumber}`,
                "journalTypeId": salesJournalId
            }
        };
        
        genericFetchPublicAPI(headerWithOrganization, qm.postSalesInvoice, queryVariables, handleResponse);
        function handleResponse(data) {
            if (data && data.data && data.data.postSalesInvoice) {
                displayBusinessInfo(`Sales Invoice ${entry.documentDate.split('T')[0]} ${entry.status} for customer ${entry.customer ? entry.customer.code : "unknown"} is posted`,false);
                resolve(data.data.postSalesInvoice);
            }
            else{
                resolve();
            }
        }
    });
}
