
// fetch data from the public API
//   headers to define the type of header (headerWithoutTenant, headerWithoutOrganization, headerWithOrganization)
//   qmObject.body to define the query or mutation
//   qmObject.context to define the context of the call
//   qmObject.policy to define the policy of the call
//   qmObject.isImportFromTemplatesRunning to define if the call is an import from templates
//   qmObject.noDataMessage to define a customized message to display when no data is found
//   variables to define the variables used by the query or mutation
//   functionCallback to define the function to call after the fetch
//   whereClause to define a dynamic where clause for the query that will replace the joker %%WHERECLAUSE%% in the query
function genericFetchPublicAPI(headers, qmObject, variables, functionCallback, whereClause="") {
  let retryAttempt = false;
  let refreshTokenAttempts = 0;
  waitInfo();

  var query = qmObject.body
    .replace('%%WHERECLAUSE%%', whereClause)
    .replace(/\n/g, '')
    .replace(/\s\s+/g, ' ');

  var context = qmObject.context;

  const fetchAPI = () => {
    return fetch(urlAPI, {
      method: "POST",
      headers: headers,
      body: JSON.stringify({ query: query, variables: variables })
    })
    .then(response => {
       if (response.status === 401 || (response.status === 500 && qmObject.policy === "userProfile")) {
         return response.json().then(errorBody => {
          if (response.status === 401 && errorBody.error.includes("invalid subscription key")) {
            displayBusinessError(errorBody.error, context);
          } else {
             if (refreshTokenAttempts < 5) {
              refreshTokenAttempts++;
              return refreshToken().then(newAccessToken => {
                headers['Authorization'] = `Bearer ${newAccessToken}`;
                return fetchAPI();
              });
            } else {
                displayBusinessError(response.status, context);
            }
          }
        });
      }
      else if (response.status >= 400) {
        return response.json().then(data => {
          if (qmObject.isImportFromTemplatesRunning) {
            functionCallback(response); 
          } else {
            if (data && data.errors && Array.isArray(data.errors) && data.errors.length > 0) {
              displayBusinessError(response.status+" "+data.errors[0].message || 'An unknown error occurred', context);
            } else {
              displayBusinessError(response.status, context);
            }
          }
          return null;
        });
      }
          
      return response.json();
    })
    .then(data => {
        if (data && data.errors && Array.isArray(data.errors) && data.errors.length > 0) {
            const errorCode = data.errors[0].extensions && data.errors[0].extensions.code;
            if (errorCode === "AUTH_NOT_AUTHORIZED") {
                displayBusinessError("403", context);
                return null;
            }
        displayBusinessWarning(data.errors[0].message, (qmObject.isImportFromTemplatesRunning?"[HIDE]":"")+context);
        functionCallback(data);
      } else if (data && data.data && hasTotalCountAndIsZero(data.data)) {
        waitInfo(false);
        var message = (qmObject.noDataMessage)?qmObject.noDataMessage:"NODATAFOUND";
        if (message.startsWith("[INFO]")) {
          displayBusinessInfo(message.substring(6), (qmObject.isImportFromTemplatesRunning?"[HIDE]":"")+context);
        } else {
          displayBusinessWarning(message, (qmObject.isImportFromTemplatesRunning?"[HIDE]":"")+context);
        }
        qmObject.noDataMessage = null;
        functionCallback(data);
      } else if (data) {
        functionCallback(data);
      }
    }).catch(error => {
      displayBusinessError(error.message, context);
      waitInfo(false);
    });
  };

  return fetchAPI().catch(error => {
      console.error('Error during fetch:', error.message);
  });
  
  // to know if a query has returned 0 records
  function hasTotalCountAndIsZero(obj) {   
    for (let key in obj) {
      if (obj.hasOwnProperty(key)) {
        if (key === "totalCount" && obj[key] === 0) return true;
        if (typeof obj[key] === 'object') {
          if (hasTotalCountAndIsZero(obj[key])) return true;
        }
      }
    }
    return false;
  }
    
} 
